/*
 * Support for executing code in new tasks and threads.
 *
 * Copyright (C) 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/sys/err.h>
#include <l4/sys/factory.h>
#include <l4/sys/task.h>
#include <l4/sys/thread.h>
#include <l4/util/elf.h>
#include <l4/util/util.h>

#include <l4/libloader/adjust_stack>

#include <fsclient/client.h>
#include <ipc/cap_alloc.h>
#include <ipc/mem_ipc.h>
#include <ipc/server.h>
#include <mem/memory_utils.h>
#include <systypes/fcntl.h>

#include <map>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pthread-l4.h>
#include <pthread.h>

#include "pager_object_interface.h"
#include "pager_object_server.h"



/* Region data structures. */

class Region
{
public:
  l4_addr_t start;
  unsigned int log2size;
  l4_umword_t flags;
  l4_addr_t map_start;

  explicit Region()
  : start(0), log2size(0), flags(0), map_start(0)
  {
  }

  explicit Region(l4_addr_t start, unsigned int log2size, l4_umword_t flags,
                  l4_addr_t map_start)
  : start(start), log2size(log2size), flags(flags), map_start(map_start)
  {
  }
};



/* Program segment abstraction. */

class Segment
{
protected:
  Region _region;

public:
  /* Allocated memory. */

  char *buf;
  l4re_ds_t ds;

  /* Segment base and corresponding region base. */

  l4_addr_t base, region_base;

  /* Segment size and corresponding region size. */

  offset_t size, region_size;

  /* Offset of segment content within the region. */

  offset_t region_offset;

  /* Access flags. */

  l4re_rm_flags_t flags;

  /* File access details. */

  offset_t file_offset, file_contents;

  explicit Segment(offset_t base, offset_t size, l4re_rm_flags_t flags,
                   offset_t file_offset = 0, offset_t file_contents = 0)
  : base(base), size(size), flags(flags), file_offset(file_offset), file_contents(file_contents)
  {
    region_base = trunc(base, L4_PAGESIZE);
    region_offset = base - region_base;
    region_size = round(size, L4_PAGESIZE);
  }

  long allocate();

  long fill(file_t *file);

  Region &region();

  l4_addr_t region_address(char *address);

  l4_addr_t region_address(l4_addr_t address);
};

long Segment::allocate()
{
  return ipc_allocate_align(size, L4RE_RM_F_SEARCH_ADDR | flags,
                           L4_PAGESHIFT, (void **) &buf, &ds);
}

long Segment::fill(file_t *file)
{
  if (!file_contents)
    return L4_EOK;

  memset(buf, 0, region_size);
  client_seek(file, file_offset, SEEK_SET);
  offset_t nread = client_read(file, buf + region_offset, file_contents);

  if (nread < file_contents)
    return -L4_EIO;
  else
    return L4_EOK;
}

Region &Segment::region()
{
  _region = Region((l4_addr_t) buf, page_order(region_size), flags, region_base);
  return _region;
}

l4_addr_t Segment::region_address(char *address)
{
  return (l4_addr_t) ((address - buf) + (char *) region_base);
}

l4_addr_t Segment::region_address(l4_addr_t address)
{
  return (address - (l4_addr_t) buf) + region_base;
}



/* A simple system pager also acting as a region mapper. */

typedef std::map<l4_addr_t, Region> Regions;

class ExecPager : public PagerObject
{
protected:
  Regions _regions;

public:
  virtual void add(Region region)
  {
    _regions[region.map_start] = region;
  }

  /* Notification methods. */

  virtual long exception(l4_exc_regs_t regs,
                         l4_snd_fpage_t *region);

  virtual long page_fault(l4_umword_t pfa, l4_umword_t pc,
                          l4_snd_fpage_t *region);

  /* Region manager/mapper methods. */

  virtual long attach(address_t *start, offset_t size, map_flags_t flags,
                      l4_cap_idx_t ds, address_t offset, unsigned char align);

};

/* Handle a general exception. */

long ExecPager::exception(l4_exc_regs_t regs, l4_snd_fpage_t *region)
{
  (void) region;

  printf("exception(...) -> pfa = %lx, pc = %lx\n", l4_utcb_exc_pfa(&regs), l4_utcb_exc_pc(&regs));

  printf("r15 = %lx\n", regs.r15);
  printf("r14 = %lx\n", regs.r14);
  printf("r13 = %lx\n", regs.r13);
  printf("r12 = %lx\n", regs.r12);
  printf("r11 = %lx\n", regs.r11);
  printf("r10 = %lx\n", regs.r10);
  printf("r9 = %lx\n", regs.r9);
  printf("r8 = %lx\n", regs.r8);
  printf("rdi = %lx\n", regs.rdi);
  printf("rsi = %lx\n", regs.rsi);
  printf("rbp = %lx\n", regs.rbp);
  printf("pfa = %lx\n", regs.pfa);
  printf("rbx = %lx\n", regs.rbx);
  printf("rdx = %lx\n", regs.rdx);
  printf("rcx = %lx\n", regs.rcx);
  printf("rax = %lx\n", regs.rax);
  printf("trapno = %lx\n", regs.trapno);
  printf("err = %lx\n", regs.err);
  printf("ip = %lx\n", regs.ip);
  printf("flags = %lx\n", regs.flags);
  printf("sp = %lx\n", regs.sp);
  printf("ss = %lx\n", regs.ss);
  printf("fs_base = %lx\n", regs.fs_base);
  printf("gs_base = %lx\n", regs.gs_base);

  return L4_EOK;
}

#define DEBUG 0

/* Handle a page fault using any configured regions. */

long ExecPager::page_fault(l4_umword_t pfa, l4_umword_t pc, l4_snd_fpage_t *region)
{
  l4_umword_t addr = pfa & ~7UL, flags = pfa & 7;

#if DEBUG
  printf("page_fault(%lx, %lx) -> %lx (%lx) -> ", pfa, pc, addr, flags);
#endif

  Regions::iterator it = _regions.upper_bound(addr);

  if (it != _regions.begin())
    it--;
  else
  {
    printf("not mapped!\n");
    return -L4_ENOMEM;
  }

  Region &r = it->second;

  if ((addr >= r.map_start) && (addr < r.map_start + (1UL << r.log2size)))
  {
    l4_addr_t page_addr = trunc(addr, L4_PAGESIZE);

    region->fpage = l4_fpage(r.start + (page_addr - r.map_start), L4_PAGESHIFT, r.flags);
    region->snd_base = page_addr;

#if DEBUG
    printf("%lx...%lx from %lx...%lx size %d rights %x\n",
           r.map_start, region->snd_base,
           r.start, l4_fpage_memaddr(region->fpage),
           l4_fpage_size(region->fpage),
           l4_fpage_rights(region->fpage));
    printf("%lx -> ", addr);

    for (unsigned int i = 0; i < sizeof(l4_umword_t); i++)
      printf("%02x", *((unsigned char *)(r.start + (addr - r.map_start) + i)));

    printf("\n");
#endif

    return L4_EOK;
  }

#if DEBUG
  printf("not mapped!\n");
#endif

  return -L4_ENOMEM;
}

/* Attach a region for provision when page faults occur. This is required in
   the initialisation of a program by the C library which requires a region
   mapper. */

long ExecPager::attach(address_t *start, offset_t size, map_flags_t flags,
                       l4_cap_idx_t ds, address_t offset, unsigned char align)
{
#if DEBUG
  printf("attach(%lx, %ld, %lx, ..., %lx, %d)\n", *start, size, flags, offset, align);
#endif

  if (align < L4_PAGESHIFT)
    align = L4_PAGESHIFT;

  offset_t increment = 1UL << align;
  offset_t region_size = round(size, increment);

  /* Either attempt to find an address for the specified region, starting from
     any indicated address. */

  if (flags & L4RE_RM_F_SEARCH_ADDR)
  {
    address_t region_start = trunc(*start, increment);
    Regions::iterator it = _regions.upper_bound(*start);

    if (!region_start)
      region_start += increment;

#if DEBUG
    printf("-> search from %lx -> %lx...\n", *start, region_start);
#endif

    /* Before last known region. */

    while (it != _regions.end())
    {
      Regions::iterator next = it;
      Region &r = it->second;
      address_t start_limit;
      address_t end_limit = r.map_start;

      /* Consider any preceding region. If no such region exists, choose an
         address at the start of memory. */

      if (it == _regions.begin())
        start_limit = L4_PAGESIZE;
      else
      {
        it--;
        Region &pr = it->second;
        start_limit = pr.map_start + (1UL << pr.log2size);
        it = next;
      }

      /* Test against the limits. */

      if (region_start < start_limit)
        region_start = round(start_limit, increment);

      /* Investigate subsequent regions if not enough space exists between the
         preceding region (or start of memory) and the current region. */

      if ((region_start + region_size) > end_limit)
      {
        it++;
        if (it == _regions.end())
          return -L4_ENOMEM;
      }
      else
        break;
    }

    /* Attach the provided dataspace.
       NOTE: This is only done in this implementation to support the paging
             mechanism. In a region mapper residing within the actual task, the
             dataspace's map operation would be invoked to obtain mappings. */

    l4_addr_t ds_start;

    long err = ipc_attach_dataspace(ds, size, (void **) &ds_start);

    if (err)
      return err;

    l4_touch_rw((const void *) ds_start, size);

#if DEBUG
    printf("-> added region for %lx size %ld (%d)\n", region_start, region_size, page_order(region_size));
#endif

    add(Region(ds_start, page_order(region_size), flags & L4RE_DS_F_RIGHTS_MASK, region_start));

    *start = region_start;
    return L4_EOK;
  }

  /* Or attempt to add the specified region at a specific address. */

  else
  {
    // NOTE: To be implemented.

    return -L4_ENOMEM;
  }
}



static ExecPager exec_pager;

static void init_pager(ipc_server_config_type *config)
{
  ipc_server_init_config(config);

  config->expected_items = PagerObject_expected_items;
  config->handler = (ipc_server_handler_type) handle_PagerObject;
  config->handler_obj = static_cast<PagerObject *>(&exec_pager);
}

static long start_pager(ipc_server_config_type *config, pthread_t thread)
{
  config->config_thread = 1;
  config->thread = pthread_l4_cap(thread);

  printf("Starting pager thread...\n");
  return ipc_server_start_config(config);
}



/* UTCB properties.
   See: moe/server/src/loader.cc */

enum ipc_exec_utcb
{
  Default_max_threads = 16,
#ifdef ARCH_mips
  Utcb_area_start = 0x73000000,
#else
  Utcb_area_start = 0xb3000000,
#endif
};



/* Capability mapping definitions for the new task. */

struct mapped_cap
{
  l4_cap_idx_t cap;
  unsigned char rights;
  l4_umword_t spot;
};

static long map_capabilities(l4_cap_idx_t task, struct mapped_cap mapped_caps[])
{
  long err = L4_EOK;
  int i = 0;

  while (l4_is_valid_cap(mapped_caps[i].cap) && !err)
  {
    err = l4_error(l4_task_map(task, L4RE_THIS_TASK_CAP,
                               l4_obj_fpage(mapped_caps[i].cap, 0, mapped_caps[i].rights),
                               l4_map_obj_control(mapped_caps[i].spot, L4_MAP_ITEM_MAP)));
    i++;
  }

  return err;
}



int main(int argc, char *argv[])
{
  long err;

  if (argc < 2)
  {
    printf("Need a program to run.\n");
    return 1;
  }

  /* Allocate capabilities for the task and thread. */

  l4_cap_idx_t caps[2];

  err = ipc_cap_alloc_many(caps, 2);

  if (err)
  {
    printf("Could not allocate capabilities.\n");
    return 1;
  }

  l4_cap_idx_t &task = caps[0];
  l4_cap_idx_t &thread = caps[1];

  /* Obtain the payload as a dataspace. */

  file_t *file = client_open(argv[1], O_RDONLY);

  if (file == NULL)
  {
    printf("Could not read file: %s\n", argv[1]);
    return 1;
  }

  /* Copy the payload regions to new dataspaces.
     NOTE: This should be directed by the ELF metadata. */

  address_t program_start = 0x1000af3;

  Segment program(0x1000000, 0x282ae, L4_FPAGE_RWX, 0, 0x282ae);
  Segment data(0x1029360, 0x8068, L4_FPAGE_RW, 0x28360, 0x2058);
  Segment stack(0x80000000, 16 * L4_PAGESIZE, L4_FPAGE_RW);
  Segment *segments[] = {&program, &data, &stack, NULL};
  Segment **segment;

  for (segment = segments; *segment != NULL; segment++)
  {
    err = (*segment)->allocate();

    if (err)
    {
      printf("Could not reserve memory.\n");
      return 1;
    }

    err = (*segment)->fill(file);

    if (err)
    {
      printf("Could not fill segment from file.\n");
      return 1;
    }
  }

  if (memcmp(program.buf + program_start - program.region_base, "\x31\xed", 2))
  {
    printf("Did not find expected instructions at start.\n");
    return 1;
  }

  /* UTCB location and size. */

  l4_addr_t utcb_start = Utcb_area_start;
  int utcb_log2size = page_order(Default_max_threads * L4_UTCB_OFFSET);

  /* Round up to at least one page. */

  if (utcb_log2size < L4_PAGESHIFT)
    utcb_log2size = L4_PAGESHIFT;

  l4_fpage_t utcb_fpage = l4_fpage(utcb_start, utcb_log2size, 0);

  /* KIP allocation. */

  l4_addr_t kip_start = (l4_addr_t) l4re_kip();

  printf("KIP at %lx.\n", kip_start);

  /* Populate stack with additional capabilities. */

  l4re_env_cap_entry_t *stack_env_cap = (l4re_env_cap_entry_t *) (stack.buf + stack.size);

  /* Special invalid/terminating environment capability entry. */

  *(--stack_env_cap) = l4re_env_cap_entry_t();

  printf("Stack region end: %p\n", stack_env_cap);

  l4_addr_t caps_start = (l4_addr_t) stack_env_cap;
  l4_umword_t *stack_element = (l4_umword_t *) stack_env_cap;

  /* Populate stack with argument values. */

  char *stack_arg = (char *) stack_element;

  stack_arg = (char *) trunc((offset_t) stack_arg - strlen(argv[1]) - 1, sizeof(l4_umword_t));

  memset(stack_arg, 0, (char *) stack_element - stack_arg);
  memcpy(stack_arg, argv[1], strlen(argv[1]));

  printf("Stack L4 program argument: %p / %lx\n", stack_arg, stack.region_address(stack_arg));

  stack_element = (l4_umword_t *) stack_arg;

  /* Loader flags, debugging flags, and the KIP capability index.
     See: generate_l4aux in Remote_app_model */

  *(--stack_element) = 0;
  *(--stack_element) = 0;
  *(--stack_element) = 0x14 << L4_CAP_SHIFT;

  printf("Stack L4 aux elements: %p / %lx\n", stack_element, stack.region_address((char *) stack_element));

  /* Populate stack with standard capabilities. */

  l4re_env_t *env = (l4re_env_t *) stack_element;

  env--;
  env->factory          = L4_BASE_FACTORY_CAP;
  env->main_thread      = L4_BASE_THREAD_CAP;
  env->log              = L4_BASE_LOG_CAP;
  env->scheduler        = L4_BASE_SCHEDULER_CAP;
  env->rm               = 0x11 << L4_CAP_SHIFT;
  env->mem_alloc        = 0x12 << L4_CAP_SHIFT;
  env->first_free_cap   = 0x15;
  env->caps             = (l4re_env_cap_entry_t *) (stack.region_address(caps_start));
  env->utcb_area        = utcb_fpage;
  env->first_free_utcb  = utcb_start + L4_UTCB_OFFSET;

  stack_element = (l4_umword_t *) env;

  /* Populate stack with AUXV. */

  /* AUXV NULL. */

  *(--stack_element) = 0;
  *(--stack_element) = 0;

  /* L4Re global environment pointer. */

  *(--stack_element) = stack.region_address((char *) env);
  *(--stack_element) = 0xf1;

  /* Apparently required entries. */

  *(--stack_element) = L4_PAGESIZE;
  *(--stack_element) = AT_PAGESZ;

  *(--stack_element) = 0;
  *(--stack_element) = AT_UID;

  *(--stack_element) = 0;
  *(--stack_element) = AT_EUID;

  *(--stack_element) = 0;
  *(--stack_element) = AT_GID;

  *(--stack_element) = 0;
  *(--stack_element) = AT_EGID;

  l4_addr_t stack_data_end = (l4_addr_t) stack_element;

  printf("Stack L4 data: %lx / %lx\n", stack_data_end, stack.region_address(stack_data_end));

  /* No environment pointers. */

  *(--stack_element) = 0;

  /* Populate stack with argument pointers and count. */
  /* NOTE: Just one argument currently. */

  *(--stack_element) = 0;
  *(--stack_element) = (l4_umword_t) (stack.region_address(stack_arg));
  *(--stack_element) = 1;

  char *stack_adjusted = Ldr::adjust_sp((char *) stack_element, NULL);

  /* Adjust the stack alignment. */

  if (stack_adjusted != (char *) stack_element)
    memmove(stack_adjusted, (const void *) stack_element, stack_data_end - (l4_addr_t) stack_element);

  l4_umword_t *stack_adjusted_element = (l4_umword_t *) stack_adjusted;

  printf("%ld %lx %lx\n", stack_adjusted_element[0], stack_adjusted_element[1], stack_adjusted_element[2]);

  l4_addr_t stack_start = stack.region_address(stack_adjusted);

  printf("Stack L4 start: %p / %lx\n", stack_adjusted, stack.region_address(stack_adjusted));

  /* Create a new task and thread. */

  err = l4_error(l4_factory_create_task(l4re_env()->factory, task, utcb_fpage));

  if (err)
  {
    printf("Could not create task.\n");
    return 1;
  }

  err = l4_error(l4_factory_create_thread(l4re_env()->factory, thread));

  if (err)
  {
    printf("Could not create thread.\n");
    return 1;
  }

  /* Start the pager. */

  ipc_server_config_type config;
  pthread_t pager_thread;
  pthread_attr_t attr;

  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

  init_pager(&config);

  for (segment = segments; *segment != NULL; segment++)
    exec_pager.add((*segment)->region());

  err = pthread_create(&pager_thread, &attr, ipc_server_start_mainloop, &config);

  if (err)
  {
    printf("Could not start pager thread.\n");
    return 1;
  }

  err = start_pager(&config, pager_thread);

  if (err)
  {
    printf("Could not start pager.\n");
    return 1;
  }

  /* Define capability mappings for the new task. */

  struct mapped_cap mapped_caps[] = {
    {config.server,         L4_CAP_FPAGE_RWS, 0x10 << L4_CAP_SHIFT},
    {config.server,         L4_CAP_FPAGE_RWS, 0x11 << L4_CAP_SHIFT},
    {task,                  L4_CAP_FPAGE_RWS, L4_BASE_TASK_CAP},
    {thread,                L4_CAP_FPAGE_RWS, L4_BASE_THREAD_CAP},
    {l4re_env()->factory,   L4_CAP_FPAGE_RWS, L4_BASE_FACTORY_CAP},
    {l4re_env()->log,       L4_CAP_FPAGE_RWS, L4_BASE_LOG_CAP},
    {l4re_env()->scheduler, L4_CAP_FPAGE_RWS, L4_BASE_SCHEDULER_CAP},
    {l4re_env()->mem_alloc, L4_CAP_FPAGE_RWS, 0x12 << L4_CAP_SHIFT},
    {L4_INVALID_CAP,        0,                0},
    };

  err = map_capabilities(task, mapped_caps);

  if (err)
  {
    printf("Could not capabilities into task.\n");
    return 1;
  }

  /* Map the KIP into the task. */

  err = l4_error(l4_task_map(task, L4RE_THIS_TASK_CAP,
                             l4_fpage(kip_start, L4_PAGESHIFT, L4_FPAGE_RX),
                             kip_start));

  if (err)
  {
    printf("Could not map KIP into task.\n");
    return 1;
  }

  /* Configure the thread with the region manager acting as pager and exception
     handler. The UTCB will be situated at an address supported by a dataspace
     attached to the new task. */

  printf("Configure thread...\n");

  l4_thread_control_start();
  l4_thread_control_pager(0x10 << L4_CAP_SHIFT);
  l4_thread_control_exc_handler(0x10 << L4_CAP_SHIFT);
  l4_thread_control_bind((l4_utcb_t *) utcb_start, task);
  err = l4_error(l4_thread_control_commit(thread));

  if (err)
  {
    printf("Could not configure thread.\n");
    return 1;
  }

  /* Start the new thread. */

  printf("Schedule thread...\n");

  printf("Stack at 0x%lx mapped to region at 0x%lx.\n", stack_start, stack.region_base);

  err = l4_error(l4_thread_ex_regs(thread, program_start, stack_start, 0));

  if (err)
  {
    printf("Could not set thread registers.\n");
    return 1;
  }

  printf("Run thread...\n");

  l4_sched_param_t sp = l4_sched_param(L4RE_MAIN_THREAD_PRIO);

  err = l4_error(l4_scheduler_run_thread(l4re_env()->scheduler, thread, &sp));

  if (err)
  {
    printf("Could not run thread.\n");
    return 1;
  }

  printf("Finished.\n");
  while (1);

  return 0;
}

/* vim: tabstop=2 expandtab shiftwidth=2
*/
